#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from shutil import which

from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.engines.pil import Engine
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsWithMaxWidthAndMaxHeight(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path

        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.AUTO_WEBP = True
        cfg.MAX_WIDTH = 150
        cfg.MAX_HEIGHT = 150

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)

        return ctx

    @gen_test
    async def test_should_be_ok_but_150x150(self):
        response = await self.async_fetch("/unsafe/200x200/grayscale.jpg")
        engine = Engine(self.context)
        engine.load(response.body, ".jpg")
        expect(response.code).to_equal(200)
        expect(response.headers["Content-Type"]).to_equal("image/jpeg")
        expect(engine.size).to_equal((150, 150))


class ImageOperationsWithMaxWidthAndMaxHeightShouldNotUpscale(
    BaseImagingTestCase
):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path

        cfg.MAX_WIDTH = 5000
        cfg.MAX_HEIGHT = 5000
        cfg.AUTO_WEBP = True
        cfg.MAX_AGE = 315360000
        cfg.ENGINE_THREADPOOL_SIZE = 2
        cfg.STORAGE = "thumbor.storages.no_storage"

        # Tried with optimizer on and off and made no difference
        cfg.OPTIMIZERS = [
            "thumbor.optimizers.jpegtran",
        ]

        cfg.RESULT_STORAGE = "thumbor.result_storages.no_storage"

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)

        return ctx

    @gen_test
    async def test_should_return_original_image_size(self):
        response = await self.async_fetch(
            "/unsafe/filters:format(jpeg)/very-small-jpeg.jpg"
        )

        engine = Engine(self.context)
        engine.load(response.body, ".jpg")
        expect(response.code).to_equal(200)
        expect(response.headers["Content-Type"]).to_equal("image/jpeg")
        expect(engine.size).to_equal((100, 100))

    @gen_test
    async def test_should_return_original_image_size_with_normalize(self):
        response = await self.async_fetch(
            "/unsafe/filters:format(jpeg)/weird_normalize_error.jpg"
        )

        engine = Engine(self.context)
        engine.load(response.body, ".jpg")
        expect(response.code).to_equal(200)
        expect(response.headers["Content-Type"]).to_equal("image/jpeg")
        expect(engine.size).to_equal((159, 239))


class ImageOperationsWithMaxPixels(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.MAX_PIXELS = 1000

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")
        return ctx

    @gen_test
    async def test_should_error(self):
        response = await self.async_fetch("/unsafe/200x200/grayscale.jpg")
        expect(response.code).to_equal(400)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest.mock import patch

import pytest
from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.engines.pil import Engine
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class EngineLoadException(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.FILTERS = []

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @patch.object(Engine, "load", side_effect=ValueError)
    @gen_test
    async def test_should_error_on_engine_load_exception(self, _):
        response = await self.async_fetch("/unsafe/image.jpg")
        expect(response.code).to_equal(500)

    @pytest.mark.skip(
        reason="I disagree with this test. If the engine fails we "
        "should get an error, otherwise we may end up caching an invalid crop."
    )
    @gen_test
    async def test_should_release_ioloop_on_error_on_engine_exception(self):
        response = await self.async_fetch("/unsafe/fit-in/134x134/940x2.png")
        expect(response.code).to_equal(200)

    @pytest.mark.skip(
        reason="I disagree with this test. If the engine fails we "
        "should get an error, otherwise we may end up caching an invalid crop."
    )
    @gen_test
    async def test_should_exec_other_operations_on_error_on_engine_exception(
        self,
    ):  # NOQA
        response = await self.async_fetch(
            "/unsafe/fit-in/134x134/filters:equalize()/940x2.png"
        )
        expect(response.code).to_equal(200)

    @patch.object(Engine, "read", side_effect=Exception)
    @gen_test
    async def test_should_fail_with_500_upon_engine_read_exception(
        self, _
    ):  # NOQA
        response = await self.async_fetch("/unsafe/fit-in/134x134/940x2.png")
        expect(response.code).to_equal(500)



import struct
from io import BytesIO

PREFIXES = [
    b"MM\x00\x2A",  # Valid TIFF header with big-endian byte order
    b"II\x2A\x00",  # Valid TIFF header with little-endian byte order
    b"MM\x2A\x00",  # Invalid TIFF header, assume big-endian
    b"II\x00\x2A",  # Invalid TIFF header, assume little-endian
]

EXIF_HEADER = b"Exif\x00\x00"


class ExifOrientationEditor:
    _endian = "<"
    _offset = None

    def __init__(self, data):
        if data[:6] != EXIF_HEADER:
            raise SyntaxError(f"not a TIFF file (header {data[:4]} not valid)")

        # Skip 6 bytes of exif header to simplify seeks in ImageFileDirectory
        self.exif_buffer = BytesIO(data[6:])

        header = self._read_header()
        self._find_orientation_offset(header)

    def _read_header(self):
        header = self.exif_buffer.read(8)

        if header[:4] not in PREFIXES:
            raise SyntaxError(f"not a TIFF file (header {header} not valid)")
        prefix = header[:2]
        if prefix == b"MM":
            self._endian = ">"
        elif prefix == b"II":
            self._endian = "<"
        else:
            raise SyntaxError("not a TIFF IFD")

        return header

    def _find_orientation_offset(self, header):
        (ifd_offset,) = self._unpack("L", header[4:])
        self.exif_buffer.seek(ifd_offset)

        # Read tag directory
        for _ in range(self._unpack("H", self.exif_buffer.read(2))[0]):
            # Each tag is 12 bytes. HHL4s = tag, type, count, data
            # Read tag and ignore the rest
            (tag,) = self._unpack("H10x", self.exif_buffer.read(12))
            if tag == 0x0112:  # Orientation tag
                self._offset = (
                    self.exif_buffer.tell() - 4
                )  # Back 4 bytes to the start of data
                break

    def _unpack(self, fmt, data):
        return struct.unpack(self._endian + fmt, data)

    def get_orientation(self):
        if self._offset is None:
            return None

        self.exif_buffer.seek(self._offset)
        return self._unpack("H", self.exif_buffer.read(2))[0]

    def set_orientation(self, value):
        if self._offset is None:
            return

        self.exif_buffer.seek(self._offset)
        self.exif_buffer.write(struct.pack(self._endian + "H", value))

    def tobytes(self):
        self.exif_buffer.seek(0)
        return EXIF_HEADER + self.exif_buffer.read()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import argparse

from thumbor import __release_date__, __version__
from thumbor.context import ServerParameters


def get_server_parameters(arguments=None):
    if arguments is None:
        arguments = []

    parser = argparse.ArgumentParser(description="thumbor server")

    parser.add_argument(
        "--version",
        action="version",
        version=f"Thumbor v{__version__} ({__release_date__})",
    )

    parser.add_argument(
        "-p",
        "--port",
        default=8888,
        type=int,
        help="The port to run this thumbor instance at [default: %(default)s].",
    )

    parser.add_argument(
        "-i",
        "--ip",
        default="0.0.0.0",
        help="The host address to run this thumbor instance at [default: %(default)s].",
    )

    parser.add_argument(
        "-f",
        "--fd",
        help="The file descriptor number or path to unix socket to listen for connections "
        "on (--port and --ip will be ignored if this is set)"
        "[default: %(default)s].",
    )

    parser.add_argument(
        "-c",
        "--conf",
        default=None,
        help="The path of the configuration file to use for this "
        "thumbor instance [default: %(default)s].",
    )

    parser.add_argument(
        "-k",
        "--keyfile",
        default=None,
        help="The path of the configuration file to use for this "
        "thumbor instance [default: %(default)s].",
    )

    parser.add_argument(
        "-l",
        "--log-level",
        default="warning",
        help="The log level to be used. Possible values are: "
        "debug, info, warning, error, critical or notset. "
        "[default: %(default)s].",
    )

    parser.add_argument(
        "-a",
        "--app",
        default="thumbor.app.ThumborServiceApp",
        help="A custom app to use for this thumbor server in case "
        "you subclassed ThumborServiceApp [default: %(default)s].",
    )

    parser.add_argument(
        "-d",
        "--debug",
        default=False,
        action="store_true",
        help="Debug mode [default: %(default)s].",
    )

    parser.add_argument(
        "--use-environment",
        default=False,
        help="Use environment variables for config",
    )

    parser.add_argument(
        "--processes",
        default=1,
        type=int,
        help="Number of processes to run. By default 1 and means no forks created"
        "Set to 0 to detect the number of cores available on this machine"
        "Set > 1 to start that specified number of processes"
        "[default: %(default)s].",
    )

    options = parser.parse_args(arguments)

    return ServerParameters(
        port=options.port,
        ip=options.ip,
        config_path=options.conf,
        keyfile=options.keyfile,
        log_level=options.log_level,
        app_class=options.app,
        debug=options.debug,
        fd=options.fd,
        use_environment=options.use_environment,
        processes=options.processes,
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from importlib import import_module

from thumbor.utils import logger


def import_class(name, get_module=False):
    module_name = name if get_module else ".".join(name.split(".")[:-1])
    klass = name.split(".")[-1]

    module = import_module(module_name)
    return module if get_module else getattr(module, klass)


def noop(func):
    def noop_decorator(*args, **kw):
        return func(*args, **kw)  # pass through

    return noop_decorator


# Unfortunately this is a very useful class and rearchitecting it is going to be hard
# So while we agree that it has too many attributes, it will remain like that for now
class Importer:  # pylint: disable=too-many-instance-attributes
    def __init__(self, config):
        self.config = config
        self.engine = None
        self.gif_engine = None
        self.loader = None
        self.url_signer = None
        self.upload_photo_storage = None
        self.storage = None
        self.metrics = None
        self.result_storage = None
        self.detectors = []
        self.filters = []
        self.optimizers = []
        self.handler_lists = []
        self.error_handler_module = None
        self.error_handler_class = None
        self.compatibility_legacy_loader = None
        self.compatibility_legacy_storage = None
        self.compatibility_legacy_result_storage = None

    @staticmethod
    def import_class(name, get_module=False, validate_fn=None):
        kls = import_class(name, get_module)
        if validate_fn is not None:
            validate_fn(kls)
        return kls

    def import_modules(self):
        if (
            self.config.COMPATIBILITY_LEGACY_LOADER
            or self.config.COMPATIBILITY_LEGACY_STORAGE
            or self.config.COMPATIBILITY_LEGACY_RESULT_STORAGE
        ):
            Importer.deprecated_monkey_patch_tornado_return_future()

        self.config.validates_presence_of(
            "ENGINE",
            "GIF_ENGINE",
            "LOADER",
            "STORAGE",
            "DETECTORS",
            "FILTERS",
            "URL_SIGNER",
            "METRICS",
            "HANDLER_LISTS",
        )

        self.import_item("ENGINE", "Engine")
        self.import_item("GIF_ENGINE", "Engine")
        self.import_item("LOADER")
        self.import_item("STORAGE", "Storage")
        self.import_item("METRICS", "Metrics")
        self.import_item("DETECTORS", "Detector", is_multiple=True)
        self.import_item(
            "FILTERS", "Filter", is_multiple=True, ignore_errors=True
        )
        self.import_item("HANDLER_LISTS", is_multiple=True)
        self.import_item("OPTIMIZERS", "Optimizer", is_multiple=True)
        self.import_item("URL_SIGNER", "UrlSigner")

        if self.config.RESULT_STORAGE:
            self.import_item("RESULT_STORAGE", "Storage")

        if self.config.UPLOAD_PHOTO_STORAGE:
            self.import_item("UPLOAD_PHOTO_STORAGE", "Storage")

        if self.config.USE_CUSTOM_ERROR_HANDLING:
            self.import_item("ERROR_HANDLER_MODULE", "ErrorHandler")
            self.error_handler_class = self.error_handler_module

        if self.config.COMPATIBILITY_LEGACY_LOADER:
            self.import_item("COMPATIBILITY_LEGACY_LOADER")

        if self.config.COMPATIBILITY_LEGACY_STORAGE:
            self.import_item("COMPATIBILITY_LEGACY_STORAGE", "Storage")

        if self.config.COMPATIBILITY_LEGACY_RESULT_STORAGE:
            self.import_item("COMPATIBILITY_LEGACY_RESULT_STORAGE", "Storage")

    @staticmethod
    def deprecated_monkey_patch_tornado_return_future():
        import tornado.concurrent  # pylint: disable=import-outside-toplevel

        if not hasattr(tornado.concurrent, "return_future"):
            setattr(tornado.concurrent, "return_future", noop)

    def import_item(
        self,
        config_key=None,
        class_name=None,
        is_multiple=False,
        item_value=None,
        ignore_errors=False,
        validate_fn=None,
    ):
        if item_value is None:
            conf_value = getattr(self.config, config_key)
        else:
            conf_value = item_value

        if is_multiple:
            self.load_multiple_item(
                config_key,
                conf_value,
                class_name,
                ignore_errors,
                validate_fn,
            )
        else:
            if class_name is not None:
                module = self.import_class(f"{conf_value}.{class_name}")
            else:
                module = self.import_class(conf_value, get_module=True)
            setattr(self, config_key.lower(), module)

    def load_multiple_item(
        self,
        config_key,
        conf_value,
        class_name,
        ignore_errors,
        validate_fn,
    ):
        modules = []
        if conf_value:
            for module_name in conf_value:
                try:
                    if class_name is not None:
                        if class_name == "*":
                            module = self.import_class(
                                module_name, validate_fn=validate_fn
                            )
                        else:
                            module = self.import_class(
                                f"{module_name}.{class_name}",
                                validate_fn=validate_fn,
                            )
                    else:
                        module = self.import_class(
                            module_name,
                            get_module=True,
                            validate_fn=validate_fn,
                        )
                    modules.append(module)
                except ImportError as error:
                    if ignore_errors:
                        logger.warning(
                            "Module %s could not be imported: %s",
                            module_name,
                            error,
                        )
                    else:
                        raise
        setattr(self, config_key.lower(), tuple(modules))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import asyncio


async def compatibility_get(*args, **kw):
    func = kw.pop("func", None)
    if func is None:
        raise RuntimeError(
            "'func' argument can't be None when calling thumbor's compatibility layer."
        )

    loop = asyncio.get_event_loop()
    queue = asyncio.Queue()

    def put(*args):
        loop.call_soon_threadsafe(queue.put_nowait, args)

    func(*args, **kw, callback=put)

    result = await queue.get()
    if isinstance(result, (tuple, list)) and len(result) == 1:
        result = result[0]

    return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

# pylint: disable=too-many-lines,consider-using-f-string

from thumbor.config import Config
from thumbor.context import Context, RequestParameters
from thumbor.detectors import BaseDetector
from thumbor.importer import Importer
from thumbor.point import FocalPoint as fp
from thumbor.storages.no_storage import Storage as NoStorage
from thumbor.threadpool import ThreadPool


def ensure(condition, message_fn):
    if not condition:
        raise AssertionError(message_fn())


class MockEngine:
    def __init__(self, size):
        self.size = size
        self.calls = {
            "resize": [],
            "crop": [],
            "vertical_flip": 0,
            "horizontal_flip": 0,
            "cover": 0,
            "reorientate": 0,
        }
        self.focal_points = None

    def reorientate(self):
        self.calls["reorientate"] += 1

    def resize(self, width, height):
        self.calls["resize"].append({"width": width, "height": height})

    def crop(self, left, top, right, bottom):
        self.calls["crop"].append(
            {"left": left, "top": top, "right": right, "bottom": bottom}
        )

    def flip_horizontally(self):
        self.calls["horizontal_flip"] += 1

    def flip_vertically(self):
        self.calls["vertical_flip"] += 1

    def get_proportional_width(self, new_height):
        width, height = self.size
        return float(new_height) * width / height

    def get_proportional_height(self, new_width):
        width, height = self.size
        return float(new_width) * height / width

    def focus(self, focal_points):
        self.focal_points = focal_points

    def extract_cover(self):
        self.calls["cover"] += 1


class MockSyncDetector(BaseDetector):
    async def detect(self):
        return []


class MockErrorSyncDetector(BaseDetector):
    async def detect(self):
        raise Exception("x")


# Test Data - pylint: disable=too-many-locals,too-many-instance-attributes
class TestData:
    def __init__(
        self,
        source_width,
        source_height,
        target_width,
        target_height,
        halign,
        valign,
        focal_points,
        crop_left,
        crop_top,
        crop_right,
        crop_bottom,
        fit_in=False,
        adaptive=False,
        full=False,
        meta=False,
        stretch=False,
    ):

        self.source_width = source_width
        self.source_height = source_height
        self.target_width = target_width
        self.target_height = target_height
        self.halign = halign
        self.valign = valign
        self.focal_points = focal_points
        self.crop_left = crop_left
        self.crop_top = crop_top
        self.crop_right = crop_right
        self.crop_bottom = crop_bottom
        self.fit_in = fit_in
        self.adaptive = adaptive
        self.full = full
        self.meta = meta
        self.stretch = stretch

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        return self.__str__()

    def __str__(self):
        crop_message = ""
        if self.crop_left is not None:
            crop_message = "it should crop %dx%d-%dx%d and " % (
                self.crop_left,
                self.crop_top,
                self.crop_right,
                self.crop_bottom,
            )
        return (
            "For an image of %dx%d resizing to %sx%s, %sit should resize to %sx%s"
        ) % (
            self.source_width,
            self.source_height,
            self.target_width,
            self.target_height,
            crop_message,
            self.target_width,
            self.target_height,
        )

    def to_context(self, detectors=None, ignore_detector_error=False):
        if detectors is None:
            detectors = []
        ThreadPool.reset()
        self.engine = MockEngine((self.source_width, self.source_height))

        flip_horizontally = flip_vertically = False

        if self.target_width != "orig":
            flip_horizontally = self.target_width < 0
            self.target_width = abs(self.target_width)

        if self.target_height != "orig":
            flip_vertically = self.target_height < 0
            self.target_height = abs(self.target_height)

        importer = Importer(None)
        importer.detectors = detectors
        importer.storage = NoStorage
        config = Config()
        config.IGNORE_SMART_ERRORS = ignore_detector_error
        ctx = Context(server=None, config=config, importer=importer)
        ctx.modules.engine = self.engine

        ctx.request = RequestParameters(
            buffer=None,
            debug=False,
            meta=self.meta,
            crop={
                "left": self.crop_left,
                "top": self.crop_top,
                "right": self.crop_right,
                "bottom": self.crop_bottom,
            },
            adaptive=self.adaptive,
            full=self.full,
            fit_in=self.fit_in,
            horizontal_flip=flip_horizontally,
            vertical_flip=flip_vertically,
            width=self.target_width,
            height=self.target_height,
            halign=self.halign,
            valign=self.valign,
            focal_points=self.focal_points,
            smart=True,
            extension="JPEG",
            filters=[],
            quality=80,
            image="some.jpeg",
            stretch=self.stretch,
        )
        ctx.request.engine = self.engine
        ctx.request.engine.extension = ".jpeg"

        return ctx

    @property
    def resize_error_message(self):
        message = "The engine resize should have been called with %sx%s" % (
            self.target_width,
            self.target_height,
        )
        if not self.engine.calls["resize"]:
            return "%s, but was never called" % message

        last_resize = self.engine.calls["resize"][0]
        return "%s, but was called with %sx%s" % (
            message,
            last_resize["width"],
            last_resize["height"],
        )

    def has_resized_properly(self):
        if not self.target_width and not self.target_height:
            return True

        same_as_source = (
            self.target_width == self.source_width
            and self.target_height == self.source_height
        )
        orig_height = (
            self.target_width == self.source_width
            and self.target_height == "orig"
        )
        orig_width = (
            self.target_width == "orig"
            and self.target_height == self.source_height
        )
        orig_both = (
            self.target_width == "orig" and self.target_height == "orig"
        )

        if same_as_source or orig_height or orig_width or orig_both:
            return True

        ensure(self.engine.calls["resize"], lambda: self.resize_error_message)

        if not self.target_width:
            ensure(
                self.engine.calls["resize"][0]["width"]
                == float(self.source_width)
                * self.target_height
                / self.source_height,
                lambda: self.resize_error_message,
            )
            ensure(
                self.engine.calls["resize"][0]["height"] == self.target_height,
                lambda: self.resize_error_message,
            )
            return True

        if not self.target_height:
            ensure(
                self.engine.calls["resize"][0]["width"] == self.target_width,
                lambda: self.resize_error_message,
            )
            ensure(
                self.engine.calls["resize"][0]["height"]
                == float(self.source_height)
                * self.target_width
                / self.source_width,
                lambda: self.resize_error_message,
            )
            return True

        ensure(
            self.engine.calls["resize"][0]["width"]
            == (
                self.target_width == "orig"
                and self.source_width
                or self.target_width
            ),
            lambda: self.resize_error_message,
        )
        ensure(
            self.engine.calls["resize"][0]["height"]
            == (
                self.target_height == "orig"
                and self.source_height
                or self.target_height
            ),
            lambda: self.resize_error_message,
        )
        return True

    @property
    def crop_error_message(self):
        message = (
            "The engine crop should have been called with %dx%d %dx%d"
            % (
                self.crop_left,
                self.crop_top,
                self.crop_right,
                self.crop_bottom,
            )
        )
        if not self.engine.calls["crop"]:
            return "%s, but was never called" % message

        last_crop = self.engine.calls["crop"][0]
        return "%s, but was called with %dx%d %dx%d" % (
            message,
            last_crop["left"],
            last_crop["top"],
            last_crop["right"],
            last_crop["bottom"],
        )

    def has_cropped_properly(self):
        if self.crop_left is None:
            ensure(
                not self.engine.calls["crop"],
                lambda: (
                    "The engine crop should NOT have been called but "
                    "was with %(left)dx%(top)d %(right)dx%(bottom)d"
                    % (self.engine.calls["crop"][0])
                ),
            )
            return True
        ensure(self.engine.calls["crop"], lambda: self.crop_error_message)
        ensure(
            (self.engine.calls["crop"][0]["left"] == self.crop_left),
            lambda: self.crop_error_message,
        )

        ensure(
            (self.engine.calls["crop"][0]["top"] == self.crop_top),
            lambda: self.crop_error_message,
        )

        ensure(
            (self.engine.calls["crop"][0]["right"] == self.crop_right),
            lambda: self.crop_error_message,
        )

        ensure(
            (self.engine.calls["crop"][0]["bottom"] == self.crop_bottom),
            lambda: self.crop_error_message,
        )

        return True


TESTITEMS = [
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=150,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=75,
        crop_right=800,
        crop_bottom=375,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=150,
        target_height=400,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=150,
        crop_top=0,
        crop_right=450,
        crop_bottom=800,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=300,
        target_height=400,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=0,
        target_height=0,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=0,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=140,
        target_width=180,
        target_height=100,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=3,
        crop_right=200,
        crop_bottom=114,
    ),
    # tests with focal points
    TestData(
        source_width=200,
        source_height=200,
        target_width=100,
        target_height=100,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 100, 1)],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=100,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 100, 1)],
        crop_left=50.0,
        crop_top=0,
        crop_right=250.0,
        crop_bottom=200,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 50, 1), fp(300, 50, 1)],
        crop_left=150.0,
        crop_top=0,
        crop_right=250.0,
        crop_bottom=200,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 150, 1), fp(300, 150, 1)],
        crop_left=150.0,
        crop_top=0,
        crop_right=250.0,
        crop_bottom=200,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 50, 1), fp(100, 150, 1)],
        crop_left=75.0,
        crop_top=0,
        crop_right=175.0,
        crop_bottom=200,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(300, 50, 1), fp(300, 150, 1)],
        crop_left=225.0,
        crop_top=0,
        crop_right=325.0,
        crop_bottom=200,
    ),
    TestData(
        source_width=200,
        source_height=400,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 50, 1), fp(300, 50, 1)],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=400,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 150, 1), fp(300, 150, 1)],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=400,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(100, 50, 1), fp(100, 150, 1)],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=400,
        target_width=100,
        target_height=200,
        halign="center",
        valign="middle",
        focal_points=[fp(300, 50, 1), fp(300, 150, 1)],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=400,
        source_height=200,
        target_width=100,
        target_height=100,
        halign="center",
        valign="middle",
        focal_points=[
            fp(50, 100, 1),
            fp(50, 300, 1),
            fp(150, 100, 1),
            fp(150, 300, 1),
        ],
        crop_left=50.0,
        crop_top=0,
        crop_right=250.0,
        crop_bottom=200,
    ),
    # Larger Width
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=300,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=275,
        crop_top=0,
        crop_right=425,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=300,
        target_height=150,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=100,
        crop_top=0,
        crop_right=700,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=300,
        target_height=150,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=600,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=275,
        crop_top=0,
        crop_right=425,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=150,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=550,
        crop_top=0,
        crop_right=700,
        crop_bottom=300,
    ),
    # Larger Height
    TestData(
        source_width=300,
        source_height=800,
        target_width=200,
        target_height=300,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=251,
        crop_right=300,
        crop_bottom=701,
    ),
    TestData(
        source_width=300,
        source_height=800,
        target_width=200,
        target_height=300,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=251,
        crop_right=300,
        crop_bottom=701,
    ),
    TestData(
        source_width=300,
        source_height=800,
        target_width=200,
        target_height=300,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=251,
        crop_right=300,
        crop_bottom=701,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=119,
        crop_right=500,
        crop_bottom=536,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=119,
        crop_right=500,
        crop_bottom=536,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=0,
        crop_top=119,
        crop_right=500,
        crop_bottom=536,
    ),
    # Proportional Height
    TestData(
        source_width=600,
        source_height=800,
        target_width=300,
        target_height=0,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=300,
        target_height=0,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=300,
        target_height=0,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=250,
        target_height=0,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=250,
        target_height=0,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=250,
        target_height=0,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    # Proportional Width
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=400,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=400,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=400,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=350,
        halign="center",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=350,
        halign="left",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=0,
        target_height=350,
        halign="right",
        valign="bottom",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=150,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=75,
        crop_right=800,
        crop_bottom=375,
    ),
    TestData(
        source_width=800,
        source_height=300,
        target_width=400,
        target_height=150,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=300,
        target_width=400,
        target_height=150,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=300,
        target_width=400,
        target_height=150,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=300,
        target_height=150,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=50,
        crop_top=0,
        crop_right=650,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=300,
        target_height=150,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=600,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=300,
        target_height=150,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=100,
        crop_top=0,
        crop_right=700,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=275,
        crop_top=0,
        crop_right=425,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=150,
        crop_bottom=300,
    ),
    TestData(
        source_width=700,
        source_height=300,
        target_width=150,
        target_height=300,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=550,
        crop_top=0,
        crop_right=700,
        crop_bottom=300,
    ),
    TestData(
        source_width=350,
        source_height=700,
        target_width=200,
        target_height=600,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=234,
        crop_bottom=700,
    ),
    TestData(
        source_width=350,
        source_height=700,
        target_width=200,
        target_height=600,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=58,
        crop_top=0,
        crop_right=292,
        crop_bottom=700,
    ),
    TestData(
        source_width=350,
        source_height=700,
        target_width=200,
        target_height=600,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=116,
        crop_top=0,
        crop_right=350,
        crop_bottom=700,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=27,
        crop_right=500,
        crop_bottom=444,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=27,
        crop_right=500,
        crop_bottom=444,
    ),
    TestData(
        source_width=500,
        source_height=600,
        target_width=300,
        target_height=250,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=0,
        crop_top=27,
        crop_right=500,
        crop_bottom=444,
    ),
    TestData(
        source_width=1,
        source_height=1,
        target_width=0,
        target_height=0,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=1,
        source_height=1,
        target_width=0,
        target_height=0,
        halign="center",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=1,
        source_height=1,
        target_width=0,
        target_height=0,
        halign="right",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=400,
        target_width=0,
        target_height=1,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=200,
        source_height=200,
        target_width=16,
        target_height=16,
        halign="left",
        valign="middle",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    # Regular
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=150,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=300,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=150,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=300,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=150,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=300,
    ),
    # Inverted
    TestData(
        source_width=600,
        source_height=800,
        target_width=150,
        target_height=400,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=300,
        crop_bottom=800,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=150,
        target_height=400,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=150,
        crop_top=0,
        crop_right=450,
        crop_bottom=800,
    ),
    TestData(
        source_width=600,
        source_height=800,
        target_width=150,
        target_height=400,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=300,
        crop_top=0,
        crop_right=600,
        crop_bottom=800,
    ),
    # Wide and Small Height
    TestData(
        source_width=800,
        source_height=60,
        target_width=400,
        target_height=15,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=30,
    ),
    TestData(
        source_width=800,
        source_height=60,
        target_width=400,
        target_height=15,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=30,
    ),
    TestData(
        source_width=800,
        source_height=60,
        target_width=400,
        target_height=15,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=800,
        crop_bottom=30,
    ),
    # Tall and Small Width
    TestData(
        source_width=60,
        source_height=800,
        target_width=15,
        target_height=400,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=30,
        crop_bottom=800,
    ),
    TestData(
        source_width=60,
        source_height=800,
        target_width=15,
        target_height=400,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=15,
        crop_top=0,
        crop_right=45,
        crop_bottom=800,
    ),
    TestData(
        source_width=60,
        source_height=800,
        target_width=15,
        target_height=400,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=30,
        crop_top=0,
        crop_right=60,
        crop_bottom=800,
    ),
    # Small Values
    TestData(
        source_width=8,
        source_height=6,
        target_width=4,
        target_height=2,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=8,
        crop_bottom=4,
    ),
    TestData(
        source_width=8,
        source_height=6,
        target_width=4,
        target_height=2,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=8,
        crop_bottom=4,
    ),
    TestData(
        source_width=8,
        source_height=6,
        target_width=4,
        target_height=2,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=8,
        crop_bottom=4,
    ),
    # Inverted Small Values
    TestData(
        source_width=6,
        source_height=8,
        target_width=2,
        target_height=4,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=0,
        crop_top=0,
        crop_right=4,
        crop_bottom=8,
    ),
    TestData(
        source_width=6,
        source_height=8,
        target_width=2,
        target_height=4,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=1,
        crop_top=0,
        crop_right=5,
        crop_bottom=8,
    ),
    TestData(
        source_width=6,
        source_height=8,
        target_width=2,
        target_height=4,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=2,
        crop_top=0,
        crop_right=6,
        crop_bottom=8,
    ),
    # Proportional Values
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=300,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=300,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=400,
        target_height=300,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    # Equal Values
    TestData(
        source_width=800,
        source_height=600,
        target_width=800,
        target_height=600,
        halign="left",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=800,
        target_height=600,
        halign="center",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=600,
        target_width=800,
        target_height=600,
        halign="right",
        valign="top",
        focal_points=[],
        crop_left=None,
        crop_top=None,
        crop_right=None,
        crop_bottom=None,
    ),
    TestData(
        source_width=800,
        source_height=400,
        target_width=400,
        target_height="orig",
        halign="middle",
        valign="middle",
        focal_points=[],
        crop_left=200,
        crop_top=0,
        crop_right=600,
        crop_bottom=400,
    ),
    TestData(
        source_width=800,
        source_height=400,
        target_width="orig",
        target_height=100,
        halign="middle",
        valign="middle",
        focal_points=[],
        crop_left=200,
        crop_top=0,
        crop_right=600,
        crop_bottom=400,
    ),
    TestData(
        source_width=800,
        source_height=400,
        target_width="orig",
        target_height="orig",
        halign="middle",
        valign="middle",
        focal_points=[],
        crop_left=200,
        crop_top=0,
        crop_right=600,
        crop_bottom=400,
    ),
]

FIT_IN_CROP_DATA = [
    (
        TestData(
            source_width=800,
            source_height=400,
            target_width=400,
            target_height=100,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (200, 100, True),
    ),
    (
        TestData(
            source_width=1000,
            source_height=250,
            target_width=500,
            target_height=200,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (500, 125, True),
    ),
    (
        TestData(
            source_width=200,
            source_height=250,
            target_width=500,
            target_height=400,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (200, 250, False),
    ),
    (
        TestData(
            source_width=800,
            source_height=400,
            target_width=100,
            target_height=400,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
            adaptive=True,
        ),
        (200, 100, True),
    ),
    (
        TestData(
            source_width=800,
            source_height=400,
            target_width=400,
            target_height=100,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
            full=True,
        ),
        (400, 200, True),
    ),
    (
        TestData(
            source_width=200,
            source_height=250,
            target_width=500,
            target_height=400,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
            full=True,
        ),
        (500, 625, True),
    ),
    (
        TestData(
            source_width=800,
            source_height=400,
            target_width=100,
            target_height=400,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
            adaptive=True,
            full=True,
        ),
        (400, 200, True),
    ),
    (
        TestData(
            source_width=500,
            source_height=500,
            target_width=250,
            target_height="orig",
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (250, 250, True),
    ),
    (
        TestData(
            source_width=500,
            source_height=500,
            target_width="orig",
            target_height=250,
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (250, 250, True),
    ),
    (
        TestData(
            source_width=500,
            source_height=500,
            target_width="orig",
            target_height="orig",
            halign="middle",
            valign="middle",
            focal_points=[],
            crop_left=None,
            crop_top=None,
            crop_right=None,
            crop_bottom=None,
            fit_in=True,
        ),
        (500, 500, False),
    ),
]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect

import thumbor.metrics
from tests.base import TestCase
from thumbor.config import Config
from thumbor.context import Context
from thumbor.importer import Importer


class StatsdMetricsTestCase(TestCase):
    def get_context(self):
        conf = Config()
        conf.METRICS = "thumbor.metrics.statsd_metrics"
        imp = Importer(conf)
        imp.import_modules()
        return Context(None, conf, imp)

    def test_should_initialize_metrics(self):
        expect(self.context.metrics).to_be_instance_of(
            thumbor.metrics.statsd_metrics.Metrics
        )

    def test_should_not_fail_on_use(self):
        expect(self.context.metrics.incr("test.count")).not_to_be_an_error()
        expect(self.context.metrics.incr("test.count", 2)).not_to_be_an_error()
        expect(
            self.context.metrics.timing("test.time", 100)
        ).not_to_be_an_error()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class BlurFilterTestCase(FilterTestCase):
    @gen_test
    async def test_blur_filter_with_sigma(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.blur", "blur(4,2)"
        )
        expected = self.get_fixture("blur.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_blur_filter_with_zero_radius(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.blur", "blur(0)"
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_equal(1)

    @gen_test
    async def test_blur_filter_without_sigma(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.blur", "blur(8)"
        )
        expected = self.get_fixture("blur2.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_blur_filter_with_max_radius(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.blur", "blur(500)"
        )
        expected = self.get_fixture("blur3.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_blur_filter_for_png_palette_mode(self):
        image = await self.get_filtered(
            "256_color_palette.png", "thumbor.filters.blur", "blur(10)"
        )
        expected = self.get_fixture("256_color_palette_blur_result.png")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_blur_filter_png_with_transparency_rgba_mode(self):
        image = await self.get_filtered(
            "PNG_transparency_demonstration_1.png",
            "thumbor.filters.blur",
            "blur(10)",
            mode="RGBA",
        )
        expected = self.get_fixture("blur4.png", mode="RGBA")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_blur_filter_png_with_transparency_la_mode(self):
        image = await self.get_filtered(
            "PNG_transparency_demonstration_1.png",
            "thumbor.filters.blur",
            "blur(10)",
            mode="LA",
        )
        expected = self.get_fixture("blur4.png", mode="LA")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from datetime import datetime, timedelta
from os.path import dirname

import pytz
from preggy import expect
from tornado.testing import gen_test

from tests.fixtures.images import default_image
from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsWithoutEtagsTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.ENABLE_ETAGS = False

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_can_get_image_without_etags(self):
        response = await self.async_fetch(
            "/unsafe/image.jpg", headers={"Accept": "image/webp,*/*;q=0.8"}
        )

        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Etag")


class ImageOperationsWithoutCorsHeaderTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.ENABLE_ETAGS = False

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_can_get_image_cors_header(self):
        response = await self.async_fetch(
            "/unsafe/image.jpg", headers={"Accept": "image/webp,*/*;q=0.8"}
        )

        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Access-Control-Allow-Origin")


class ImageOperationsWithCorsHeaderTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(
            SECURITY_KEY="ACME-SEC", ACCESS_CONTROL_ALLOW_ORIGIN_HEADER="*"
        )
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.ENABLE_ETAGS = False

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_can_get_image_with_cors_header(self):
        response = await self.async_fetch(
            "/unsafe/image.jpg", headers={"Accept": "image/webp,*/*;q=0.8"}
        )

        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Access-Control-Allow-Origin")
        expect(response.headers["Access-Control-Allow-Origin"]).to_equal("*")


class ImageOperationsWithLastModifiedTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path

        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path

        cfg.SEND_IF_MODIFIED_LAST_MODIFIED_HEADERS = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @property
    def result_storage(self):
        return self.context.modules.result_storage

    def write_image(self):
        expected_path = self.result_storage.normalize_path(
            "/_wIUeSaeHw8dricKG2MGhqu5thk=/smart/image.jpg"
        )

        if not os.path.exists(dirname(expected_path)):
            os.makedirs(dirname(expected_path))

        if not os.path.exists(expected_path):
            with open(expected_path, "wb") as img:
                img.write(default_image())

    @gen_test
    async def test_can_get_304_with_last_modified(self):
        self.write_image()
        response = await self.async_fetch(
            "/_wIUeSaeHw8dricKG2MGhqu5thk=/smart/image.jpg",
            headers={
                "Accept": "image/webp,*/*;q=0.8",
                "If-Modified-Since": (datetime.utcnow() + timedelta(seconds=1))
                .replace(tzinfo=pytz.utc)
                .strftime("%a, %d %b %Y %H:%M:%S GMT"),  # NOW +1 sec UTC
            },
        )

        expect(response.code).to_equal(304)

    @gen_test
    async def test_can_get_image_with_last_modified(self):
        self.write_image()
        response = await self.async_fetch(
            "/_wIUeSaeHw8dricKG2MGhqu5thk=/smart/image.jpg",
            headers={
                "Accept": "image/webp,*/*;q=0.8",
                "If-Modified-Since": (datetime.utcnow() - timedelta(days=365))
                .replace(tzinfo=pytz.utc)
                .strftime("%a, %d %b %Y %H:%M:%S GMT"),  # Last Year
            },
        )

        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Last-Modified")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest.mock import patch

from libthumbor import CryptoURL
from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import RequestParameters, ServerParameters
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsWithAutoPngToJpgTestCase(BaseImagingTestCase):
    def get_config(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.AUTO_PNG_TO_JPG = True
        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path
        return cfg

    def get_importer(self):
        importer = Importer(self.config)
        importer.import_modules()
        return importer

    def get_server(self):
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return server

    def get_request(self, *args, **kwargs):
        return RequestParameters(*args, **kwargs)

    def get_context(self):
        ctx = super().get_context()
        ctx.request = self.get_request()
        return ctx

    async def get_as_webp(self, url):
        return await self.async_fetch(
            url, headers={"Accept": "image/webp,*/*;q=0.8"}
        )

    @gen_test
    async def test_should_auto_convert_png_to_jpg(self):
        response = await self.async_fetch(
            "/unsafe/Giunchedi%2C_Filippo_January_2015_01.png"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_jpeg()

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_should_auto_convert_png_to_jpg_with_signed_images(
        self, context_mock
    ):
        context_mock.return_value = self.context
        crypto = CryptoURL("ACME-SEC")
        url = crypto.generate(
            image_url="Giunchedi%2C_Filippo_January_2015_01.png"
        )
        self.context.request = self.get_request(url=url)

        context_mock.return_value = self.context
        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_jpeg()

    @gen_test
    async def test_shouldnt_auto_convert_png_to_jpg_if_png_has_transparency(
        self,
    ):  # NOQA
        response = await self.async_fetch("/unsafe/watermark.png")
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_png()

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_shouldnt_auto_convert_png_to_jpg_if_png_has_transparency_with_signed_images(  # NOQA
        self, context_mock
    ):
        context_mock.return_value = self.context
        crypto = CryptoURL("ACME-SEC")
        url = crypto.generate(image_url="watermark.png")
        self.context.request = self.get_request(url=url)

        # save on result storage
        response = await self.async_fetch(url)
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_png()

    @gen_test
    async def test_should_auto_convert_png_to_webp_if_auto_webp_is_true(self):
        self.config.AUTO_WEBP = True

        response = await self.get_as_webp(
            "/unsafe/Giunchedi%2C_Filippo_January_2015_01.png"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.body).to_be_webp()

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_should_auto_convert_png_to_webp_if_auto_webp_is_true_with_signed_images(  # NOQA
        self, context_mock
    ):
        self.config.AUTO_WEBP = True
        context_mock.return_value = self.context
        crypto = CryptoURL("ACME-SEC")
        url = crypto.generate(
            image_url="Giunchedi%2C_Filippo_January_2015_01.png"
        )
        self.context.request = self.get_request(url=url, accepts_webp=True)

        # save on result storage
        response = await self.get_as_webp(url)
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_should_auto_convert_png_to_webp_if_auto_webp_is_true_and_png_has_transparency(  # NOQA
        self,
    ):
        self.config.AUTO_WEBP = True

        response = await self.get_as_webp("/unsafe/watermark.png")
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.body).to_be_webp()

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_should_auto_convert_png_to_webp_if_auto_webp_is_true_and_png_has_transparency_with_signed_images(  # NOQA
        self, context_mock
    ):
        self.config.AUTO_WEBP = True
        context_mock.return_value = self.context
        crypto = CryptoURL("ACME-SEC")
        url = crypto.generate(image_url="watermark.png")
        self.context.request = self.get_request(url=url)

        # save on result storage
        response = await self.get_as_webp(url)
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.body).to_be_webp()

    @patch("thumbor.handlers.imaging.RequestParameters")
    @gen_test
    async def test_shouldnt_auto_convert_png_to_jpg_if_requested_by_request_param(  # NOQA
        self, request_mock
    ):
        def side_effect(*_, **kwargs):
            req = RequestParameters(**kwargs)
            req.auto_png_to_jpg = False

            return req

        request_mock.side_effect = side_effect

        response = await self.async_fetch(
            "/unsafe/Giunchedi%2C_Filippo_January_2015_01.png"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_png()

    @patch("thumbor.handlers.imaging.RequestParameters")
    @gen_test
    async def test_should_auto_convert_png_to_jpg_if_requested_by_request_param(  # NOQA
        self, request_mock
    ):
        self.config.AUTO_PNG_TO_JPG = False

        def side_effect(*_, **kwargs):
            req = RequestParameters(**kwargs)
            req.auto_png_to_jpg = True

            return req

        request_mock.side_effect = side_effect

        response = await self.async_fetch(
            "/unsafe/Giunchedi%2C_Filippo_January_2015_01.png"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_jpeg()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest.mock import Mock

from preggy import expect

from tests.base import TestCase
from thumbor.handlers import FetchResult


class FetchResultTestCase(TestCase):
    def test_can_create_default_fetch_result(self):
        result = FetchResult()
        expect(result.normalized).to_be_false()
        expect(result.buffer).to_be_null()
        expect(result.engine).to_be_null()
        expect(result.successful).to_be_false()
        expect(result.loader_error).to_be_null()

    def test_can_create_fetch_result(self):
        buffer_mock = Mock()
        engine_mock = Mock()
        error_mock = Mock()
        result = FetchResult(
            normalized=True,
            buffer=buffer_mock,
            engine=engine_mock,
            successful=True,
            loader_error=error_mock,
        )
        expect(result.normalized).to_be_true()
        expect(result.buffer).to_equal(buffer_mock)
        expect(result.engine).to_equal(engine_mock)
        expect(result.successful).to_be_true()
        expect(result.loader_error).to_equal(error_mock)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor import storages
from thumbor.compatibility import compatibility_get


class Storage(storages.BaseStorage):
    @property
    def storage(self):
        storage = self.context.modules.compatibility_legacy_storage
        if storage is None:
            raise RuntimeError(
                "The 'COMPATIBILITY_LEGACY_STORAGE' configuration should point to "
                "a valid storage when using compatibility storage."
            )
        return storage

    async def put(self, path, file_bytes):
        return self.storage.put(path, file_bytes)

    async def put_crypto(self, path):
        return self.storage.put_crypto(path)

    async def put_detector_data(self, path, data):
        return self.storage.put_detector_data(path, data)

    async def get(self, path):
        return await compatibility_get(path, func=self.storage.get)

    async def get_crypto(self, path):
        return await compatibility_get(path, func=self.storage.get_crypto)

    async def get_detector_data(self, path):
        return await compatibility_get(
            path, func=self.storage.get_detector_data
        )

    async def exists(self, path, *args, **kw):
        return await compatibility_get(
            path, func=self.storage.exists, *args, **kw
        )

    async def remove(self, path):
        return self.storage.remove(path)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import unicodedata
from io import BytesIO
from os.path import exists
from unittest import skipUnless

from PIL import Image
from preggy import create_assertions

from thumbor.testing import DetectorTestCase as ThumborDetectorTestCase
from thumbor.testing import FilterTestCase as ThumborFilterTestCase
from thumbor.testing import TestCase as ThumborTestCase
from thumbor.testing import get_ssim

try:
    from PIL import _avif  # noqa pylint: disable=ungrouped-imports
except ImportError:
    try:
        from pillow_avif import _avif
    except ImportError:
        _avif = None

try:
    from pillow_heif import HeifImagePlugin
except ImportError:
    HeifImagePlugin = None

AVIF_AVAILABLE = _avif is not None
HEIF_AVAILABLE = HeifImagePlugin is not None

skip_unless_avif = skipUnless(
    AVIF_AVAILABLE,
    "AVIF format support not found. Skipping AVIF tests.",
)

skip_unless_heif = skipUnless(
    HEIF_AVAILABLE,
    "HEIF format support not found. Skipping HEIF tests.",
)


def skip_unless_avif_encoder(codec_name):
    return skipUnless(
        _avif and _avif.encoder_codec_available(codec_name),
        f"{codec_name} encode not available",
    )


class TestCase(ThumborTestCase):
    pass


class DetectorTestCase(ThumborDetectorTestCase):
    pass


class FilterTestCase(ThumborFilterTestCase):
    pass


@create_assertions
def to_exist(topic):
    return exists(topic)


def normalize_unicode_path(path):
    normalized_path = path
    for unicode_format in ["NFD", "NFC", "NFKD", "NFKC"]:
        normalized_path = unicodedata.normalize(unicode_format, str(path))
        if exists(normalized_path):
            break
    return normalized_path


@create_assertions
def to_be_the_same_as(topic, expected):
    topic = normalize_unicode_path(topic)
    expected = normalize_unicode_path(expected)

    if not exists(topic):
        raise AssertionError(f"File at {topic} does not exist")
    if not exists(expected):
        raise AssertionError(f"File at {expected} does not exist")

    topic_image = Image.open(topic)
    expected_image = Image.open(expected)

    return get_ssim(topic_image, expected_image) > 0.95


@create_assertions
def to_be_similar_to(topic, expected):
    topic_image = Image.open(BytesIO(topic))
    expected_image = Image.open(BytesIO(expected))

    return get_ssim(topic_image, expected_image) > 0.95


@create_assertions
def to_be_webp(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "webp"


@create_assertions
def to_be_avif(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "avif"


@create_assertions
def to_be_heif(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "heif"


@create_assertions
def to_be_png(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "png"


@create_assertions
def to_be_gif(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "gif"


@create_assertions
def to_be_jpeg(topic):
    image = Image.open(BytesIO(topic))
    return image.format.lower() == "jpeg"


@create_assertions
def to_be_resized(image):
    return image.has_resized_properly()


@create_assertions
def to_be_cropped(image):
    return image.has_cropped_properly()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from thumbor.doctor import run_doctor


def test_get_doctor_output(capsys, doctor_output):
    run_doctor(
        {
            "nocolor": True,
            "config": "./tests/invalid-thumbor.conf",
        },
        print_version=False,
        exit_with_error=False,
    )
    assert capsys.readouterr().out == doctor_output


def test_get_doctor_output_no_config(capsys, doctor_output_no_config):
    run_doctor(
        {
            "nocolor": True,
            "config": None,
        },
        print_version=False,
        exit_with_error=False,
    )
    assert capsys.readouterr().out == doctor_output_no_config



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from shutil import which
from unittest.mock import patch
from urllib.parse import quote

from libthumbor import CryptoURL
from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, RequestParameters, ServerParameters
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsWithAutoWebPTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.AUTO_WEBP = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")
        return ctx

    async def get_as_webp(self, url):
        return await self.async_fetch(
            url, headers={"Accept": "image/webp,*/*;q=0.8"}
        )

    @gen_test
    async def test_can_auto_convert_jpeg(self):
        response = await self.get_as_webp("/unsafe/image.jpg")
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")

        expect(response.body).to_be_webp()

    @gen_test
    async def test_should_not_convert_animated_gifs_to_webp(self):
        response = await self.get_as_webp("/unsafe/animated.gif")

        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_gif()

    @gen_test
    async def test_should_convert_image_with_small_width_and_no_height(self):
        response = await self.get_as_webp("/unsafe/0x0:1681x596/1x/image.jpg")

        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_should_convert_monochromatic_jpeg(self):
        response = await self.get_as_webp("/unsafe/grayscale.jpg")
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_should_convert_cmyk_jpeg(self):
        response = await self.get_as_webp("/unsafe/cmyk.jpg")
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_shouldnt_convert_cmyk_jpeg_if_format_specified(self):
        response = await self.get_as_webp(
            "/unsafe/filters:format(png)/cmyk.jpg"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_png()

    @gen_test
    async def test_shouldnt_convert_cmyk_jpeg_if_gif(self):
        response = await self.get_as_webp(
            "/unsafe/filters:format(gif)/cmyk.jpg"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_gif()

    @gen_test
    async def test_shouldnt_convert_if_format_specified(self):
        response = await self.get_as_webp(
            "/unsafe/filters:format(gif)/image.jpg"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_gif()

    @gen_test
    async def test_shouldnt_add_vary_if_format_specified(self):
        response = await self.get_as_webp(
            "/unsafe/filters:format(webp)/image.jpg"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).not_to_include("Vary")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_should_add_vary_if_format_invalid(self):
        response = await self.get_as_webp(
            "/unsafe/filters:format(asdf)/image.jpg"
        )
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()

    @gen_test
    async def test_converting_return_etags(self):
        response = await self.get_as_webp("/unsafe/image.jpg")
        expect(response.headers).to_include("Etag")


class ImageOperationsWithAutoWebPWithResultStorageTestCase(
    BaseImagingTestCase
):
    def get_request(self, *args, **kwargs):
        return RequestParameters(*args, **kwargs)

    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.AUTO_WEBP = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.request = self.get_request()
        ctx.server.gifsicle_path = which("gifsicle")
        return ctx

    @property
    def result_storage(self):
        return self.context.modules.result_storage

    async def get_as_webp(self, url):
        return await self.async_fetch(
            url, headers={"Accept": "image/webp,*/*;q=0.8"}
        )

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_can_auto_convert_jpeg_from_result_storage(
        self, context_mock
    ):  # NOQA
        context_mock.return_value = self.context
        crypto = CryptoURL("ACME-SEC")
        url = crypto.generate(
            image_url=quote("http://test.com/smart/image.jpg")
        )
        self.context.request = self.get_request(url=url, accepts_webp=True)
        with open("./tests/fixtures/images/image.webp", "rb") as fixture:
            await self.context.modules.result_storage.put(fixture.read())

        response = await self.get_as_webp(url)
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()

    @patch("thumbor.handlers.Context")
    @gen_test
    async def test_can_auto_convert_unsafe_jpeg_from_result_storage(
        self, context_mock
    ):
        context_mock.return_value = self.context
        self.context.request = self.get_request(accepts_webp=True)

        response = await self.get_as_webp("/unsafe/image.jpg")
        expect(response.code).to_equal(200)
        expect(response.headers).to_include("Vary")
        expect(response.headers["Vary"]).to_include("Accept")
        expect(response.body).to_be_webp()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath
from unittest import mock

from preggy import expect
from tornado.testing import gen_test

import thumbor.detectors
from tests.base import DetectorTestCase
from thumbor.detectors.glasses_detector import Detector as GlassesDetector


class GlassesDetectorTestCase(DetectorTestCase):
    @gen_test
    async def test_detector_uses_proper_cascade(self):
        cascade = "./tests/fixtures/haarcascade_eye_tree_eyeglasses.xml"
        ctx = mock.Mock(
            config=mock.Mock(
                GLASSES_DETECTOR_CASCADE_FILE=abspath(cascade),
            )
        )

        detector = GlassesDetector(ctx, 1, [])
        expect(detector).not_to_be_null()

    @gen_test
    async def test_should_detect_glasses(self):
        with open(
            abspath(
                "./tests/fixtures/images/Christophe_Henner_-_June_2016.jpg"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.GLASSES_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/glasses_detector/"
            "haarcascade_eye_tree_eyeglasses.xml",
        )

        if hasattr(GlassesDetector, "cascade"):
            del GlassesDetector.cascade
        await GlassesDetector(self.context, 0, []).detect()
        detection_result = self.context.request.focal_points[0]
        expect(detection_result.origin).to_equal("detection")
        expect(detection_result.x).to_be_numeric()
        expect(detection_result.y).to_be_numeric()
        expect(detection_result.width).to_be_numeric()
        expect(detection_result.height).to_be_numeric()

    @gen_test
    async def test_should_skip_if_opencv_not_found(self):
        with open(
            abspath(
                "./tests/fixtures/images/Christophe_Henner_-_June_2016.jpg"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.GLASSES_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/glasses_detector/"
            "haarcascade_eye_tree_eyeglasses.xml",
        )

        if hasattr(GlassesDetector, "cascade"):
            del GlassesDetector.cascade

        with mock.patch.object(
            thumbor.detectors.BaseDetector, "verify_cv", lambda self: False
        ):
            await GlassesDetector(self.context, 0, []).detect()

        detection_result = self.context.request.focal_points
        expect(detection_result).to_length(0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from typing import Any, cast

from thumbor.handler_lists import HandlerList
from thumbor.handlers.blacklist import BlacklistHandler


def get_handlers(context: Any) -> HandlerList:
    is_blacklist_enabled = cast(bool, context.config.USE_BLACKLIST)
    if not is_blacklist_enabled:
        return []

    return [(r"/blacklist/?", BlacklistHandler, {"context": context})]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.ext.filters import _nine_patch
from thumbor.filters import BaseFilter, filter_method
from thumbor.loaders import LoaderResult


class Filter(BaseFilter):
    regex = r"(?:frame\((?P<url>.*?))"

    async def on_image_ready(self, buffer):
        self.nine_patch_engine.load(buffer, None)
        self.nine_patch_engine.enable_alpha()
        self.engine.enable_alpha()

        (
            nine_patch_mode,
            nine_patch_data,
        ) = self.nine_patch_engine.image_data_as_rgb()
        padding = _nine_patch.get_padding(
            nine_patch_mode,
            nine_patch_data,
            self.nine_patch_engine.size[0],
            self.nine_patch_engine.size[1],
        )

        self.handle_padding(padding)

        mode, data = self.engine.image_data_as_rgb()

        if mode != nine_patch_mode:
            raise RuntimeError(
                f"Image mode mismatch: {mode} != {nine_patch_mode}"
            )

        imgdata = _nine_patch.apply(
            mode,
            data,
            self.engine.size[0],
            self.engine.size[1],
            nine_patch_data,
            self.nine_patch_engine.size[0],
            self.nine_patch_engine.size[1],
        )
        self.engine.set_image_data(imgdata)

    def handle_padding(self, padding):
        """Pads the image with transparent pixels if necessary."""
        left = padding[0]
        top = padding[1]
        right = padding[2]
        bottom = padding[3]

        offset_x = 0
        offset_y = 0
        new_width = self.engine.size[0]
        new_height = self.engine.size[1]

        if left > 0:
            offset_x = left
            new_width += left
        if top > 0:
            offset_y = top
            new_height += top
        if right > 0:
            new_width += right
        if bottom > 0:
            new_height += bottom
        new_engine = self.context.modules.engine.__class__(self.context)
        new_engine.image = new_engine.gen_image(
            (new_width, new_height), "#fff"
        )
        new_engine.enable_alpha()
        new_engine.paste(self.engine, (offset_x, offset_y))
        self.engine.image = new_engine.image

    async def on_fetch_done(self, result):
        # TODO if result.successful is False how can the error be handled?
        if isinstance(result, LoaderResult):
            buffer = result.buffer
        else:
            buffer = result

        self.nine_patch_engine.load(buffer, None)
        await self.storage.put(self.url, self.nine_patch_engine.read())
        await self.storage.put_crypto(self.url)
        await self.on_image_ready(buffer)

    @filter_method(BaseFilter.String)
    async def frame(self, url):
        self.url = url
        self.nine_patch_engine = self.context.modules.engine.__class__(
            self.context
        )
        self.storage = self.context.modules.storage

        buffer = await self.storage.get(self.url)
        if buffer is not None:
            return await self.on_image_ready(buffer)

        result = await self.context.modules.loader.load(self.context, self.url)
        await self.on_fetch_done(result)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import thumbor.result_storages as storages
from thumbor.compatibility import compatibility_get


class Storage(storages.BaseStorage):
    @property
    def storage(self):
        storage = self.context.modules.compatibility_legacy_result_storage
        if storage is None:
            raise RuntimeError(
                "The 'COMPATIBILITY_LEGACY_RESULT_STORAGE' configuration should point "
                "to a valid result storage when using compatibility result storage."
            )
        return storage

    async def put(self, image_bytes):
        if self.storage is None:
            return
        return self.storage.put(image_bytes)

    async def get(self):
        if self.storage is None:
            return
        return await compatibility_get(func=self.storage.get)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.loaders import http_loader


def _normalize_url(url):
    url = http_loader.quote_url(url)
    if url.startswith("http:"):
        url = url.replace("http:", "https:", 1)

    return url if url.startswith("https://") else f"https://{url}"


def validate(context, url):
    if url.startswith("http://"):
        return False

    return http_loader.validate(
        context, url, normalize_url_func=_normalize_url
    )


def return_contents(response, url, context):
    return http_loader.return_contents(response, url, context)


async def load(context, url):
    return await http_loader.load(
        context,
        url,
        normalize_url_func=_normalize_url,
        return_contents_fn=return_contents,
        encode_fn=encode,
    )


def encode(string):
    return http_loader.encode(string)



# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.ext.filters import _convolution
from thumbor.filters import BaseFilter, filter_method

# pylint: disable=line-too-long


class Filter(BaseFilter):
    """
    Usage: /filters:convolution(<semicolon separated matrix items>, <number of columns in matrix>, <should normalize boolean>)
    Example of blur filter: /filters:convolution(1;2;1;2;4;2;1;2;1,3,true)/
    """

    @filter_method(
        r"(?:[-]?[\d]+\.?[\d]*[;])*(?:[-]?[\d]+\.?[\d]*)",
        BaseFilter.PositiveNumber,
        BaseFilter.Boolean,
    )
    async def convolution(self, matrix, columns, should_normalize=True):
        matrix = tuple(matrix.split(";"))
        mode, data = self.engine.image_data_as_rgb()
        imgdata = _convolution.apply(
            mode,
            data,
            self.engine.size[0],
            self.engine.size[1],
            matrix,
            columns,
            should_normalize,
        )
        self.engine.set_image_data(imgdata)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import math
import sys

from thumbor.point import FocalPoint
from thumbor.utils import logger

TRIM_ENABLED = True
try:
    from thumbor.ext.filters import _bounding_box
except ImportError:
    logger.warning("Error importing bounding_box filter, trimming won't work")
    TRIM_ENABLED = False


class Transformer:
    def __init__(self, context):
        self.context = context
        self.engine = self.context.request.engine
        self.focal_points = None
        self.target_height = None
        self.target_width = None

    async def transform(self):
        if self.context.config.RESPECT_ORIENTATION:
            self.engine.reorientate()
        await self.trim()
        await self.smart_detect()
        await self.do_image_operations()

    async def trim(self):
        is_gifsicle = (
            self.context.request.engine.extension == ".gif"
            and self.context.config.USE_GIFSICLE_ENGINE
        )
        if (
            self.context.request.trim is None
            or not TRIM_ENABLED
            or is_gifsicle
        ):
            return

        mode, data = self.engine.image_data_as_rgb()
        box = _bounding_box.apply(
            mode,
            self.engine.size[0],
            self.engine.size[1],
            self.context.request.trim_pos,
            self.context.request.trim_tolerance,
            data,
        )

        if box[2] < box[0] or box[3] < box[1]:
            logger.warning(
                "Ignoring trim, there wouldn't be any image "
                "left, check the tolerance."
            )
            return

        self.engine.crop(box[0], box[1], box[2] + 1, box[3] + 1)
        if self.context.request.should_crop:
            self.context.request.crop["left"] -= box[0]
            self.context.request.crop["top"] -= box[1]
            self.context.request.crop["right"] -= box[0]
            self.context.request.crop["bottom"] -= box[1]

    def _calculate_target_dimensions(self):
        source_width, source_height = self.engine.size
        source_width = float(source_width)
        source_height = float(source_height)

        if not self.context.request.width and not self.context.request.height:
            self.target_width = source_width
            self.target_height = source_height
        else:
            if self.context.request.width:
                if self.context.request.width == "orig":
                    self.target_width = source_width
                else:
                    self.target_width = float(self.context.request.width)
            else:
                self.target_width = self.engine.get_proportional_width(
                    self.context.request.height
                )

            if self.context.request.height:
                if self.context.request.height == "orig":
                    self.target_height = source_height
                else:
                    self.target_height = float(self.context.request.height)
            else:
                self.target_height = self.engine.get_proportional_height(
                    self.context.request.width
                )

    def get_target_dimensions(self):
        """
        Returns the target dimensions and calculates them if necessary.
        The target dimensions are display independent.
        :return: Target dimensions as a tuple (width, height)
        :rtype: (int, int)
        """
        if self.target_height is None:
            self._calculate_target_dimensions()
        return int(self.target_width), int(self.target_height)

    def adjust_focal_points(self):
        source_width, source_height = self.engine.size

        self.focal_points = None

        if self.context.request.focal_points:
            if self.context.request.should_crop:
                self.focal_points = []
                crop = self.context.request.crop
                for point in self.context.request.focal_points:
                    if (
                        point.x < crop["left"]
                        or point.x > crop["right"]
                        or point.y < crop["top"]
                        or point.y > crop["bottom"]
                    ):
                        continue
                    point.x -= crop["left"] or 0
                    point.y -= crop["top"] or 0
                    self.focal_points.append(point)
            else:
                self.focal_points = self.context.request.focal_points

        if not self.focal_points:
            self.focal_points = [
                FocalPoint.from_alignment(
                    self.context.request.halign,
                    self.context.request.valign,
                    source_width,
                    source_height,
                )
            ]

        self.engine.focus(self.focal_points)

    @property
    def smart_storage_key(self):
        return self.context.request.image_url

    async def smart_detect(self):
        is_gifsicle = (
            self.context.request.engine.extension == ".gif"
            and self.context.config.USE_GIFSICLE_ENGINE
        )
        if (
            not (self.context.modules.detectors and self.context.request.smart)
        ) or is_gifsicle:
            return

        try:
            await self.do_smart_detection()
        except Exception:  # pylint: disable=broad-except
            if not self.context.config.IGNORE_SMART_ERRORS:
                raise

            logger.exception("Ignored error during smart detection")
            if self.context.config.USE_CUSTOM_ERROR_HANDLING:
                self.context.modules.importer.error_handler.handle_error(
                    context=self.context,
                    handler=self.context.request_handler,
                    exception=sys.exc_info(),
                )

            self.context.request.prevent_result_storage = True
            self.context.request.detection_error = True

    async def do_smart_detection(self):
        focal_points = await (
            self.context.modules.storage.get_detector_data(
                self.smart_storage_key
            )
        )
        points_from_storage = focal_points is not None
        if focal_points is None:
            detectors = self.context.modules.detectors
            focal_points = await detectors[0](
                self.context, index=0, detectors=detectors
            ).detect()

        if focal_points is not None:
            for point in focal_points:
                self.context.request.focal_points.append(
                    FocalPoint.from_dict(point)
                )

        if (
            self.context.request.focal_points
            and self.context.modules.storage
            and not points_from_storage
        ):
            storage = self.context.modules.storage
            points = []
            for point in self.context.request.focal_points:
                points.append(point.to_dict())

            await storage.put_detector_data(self.smart_storage_key, points)

    async def do_image_operations(self):
        """
        If ENGINE_THREADPOOL_SIZE > 0, this will schedule the image operations
        into a threadpool.  If not, it just executes them synchronously.

        The actual work happens in self.img_operation_worker
        """
        await self.context.thread_pool.queue(
            operation=self.img_operation_worker,
        )

    def img_operation_worker(self):
        if (
            self.context.request.engine.extension == ".gif"
            and "cover()" in self.context.request.filters
        ):
            self.extract_cover()

        self.manual_crop()
        self._calculate_target_dimensions()
        self.adjust_focal_points()

        if self.context.request.debug:
            self.debug()
        else:
            if self.context.request.fit_in:
                self.fit_in_resize()
            else:
                if not self.context.request.stretch:
                    self.auto_crop()
                self.resize()
            self.flip()

    def extract_cover(self):
        self.engine.extract_cover()

    def manual_crop(self):
        if self.context.request.should_crop:

            def limit(dimension, maximum):
                return min(max(dimension, 0), maximum)

            source_width, source_height = self.engine.size
            crop = self.context.request.crop

            crop["left"] = limit(crop["left"], source_width)
            crop["top"] = limit(crop["top"], source_height)
            crop["right"] = limit(crop["right"], source_width)
            crop["bottom"] = limit(crop["bottom"], source_height)

            if crop["left"] >= crop["right"] or crop["top"] >= crop["bottom"]:
                self.context.request.should_crop = False
                crop["left"] = crop["right"] = crop["top"] = crop["bottom"] = 0
                return

            self.engine.crop(
                crop["left"], crop["top"], crop["right"], crop["bottom"]
            )

    def auto_crop(self):
        source_width, source_height = self.engine.size

        target_height = self.target_height or 1
        target_width = self.target_width or 1

        source_ratio = round(source_width / source_height, 2)
        target_ratio = round(target_width / target_height, 2)

        if source_ratio == target_ratio:
            return

        focal_x, focal_y = self.get_center_of_mass()

        if (
            self.target_width / source_width
            > self.target_height / source_height
        ):
            crop_width = source_width
            crop_height = int(
                round(source_width * self.target_height / target_width, 0)
            )
        else:
            crop_width = int(
                round(
                    math.ceil(
                        self.target_width * source_height / target_height
                    ),
                    0,
                )
            )
            crop_height = source_height

        crop_left = int(
            round(
                min(
                    max(focal_x - (crop_width / 2), 0.0),
                    source_width - crop_width,
                )
            )
        )
        crop_right = min(crop_left + crop_width, source_width)

        crop_top = int(
            round(
                min(
                    max(focal_y - (crop_height / 2), 0.0),
                    source_height - crop_height,
                )
            )
        )
        crop_bottom = min(crop_top + crop_height, source_height)

        self.engine.crop(crop_left, crop_top, crop_right, crop_bottom)

    def flip(self):
        if self.context.request.horizontal_flip:
            self.engine.flip_horizontally()
        if self.context.request.vertical_flip:
            self.engine.flip_vertically()

    def get_center_of_mass(self):
        total_weight = 0.0
        total_x = 0.0
        total_y = 0.0

        for focal_point in self.focal_points:
            total_weight += focal_point.weight

            total_x += focal_point.x * focal_point.weight
            total_y += focal_point.y * focal_point.weight

        avg_x = total_x // total_weight
        avg_y = total_y // total_weight

        return avg_x, avg_y

    def resize(self):
        source_width, source_height = self.engine.size
        if (
            self.target_width == source_width
            and self.target_height == source_height
        ):
            return
        self.engine.resize(
            self.target_width or 1, self.target_height or 1
        )  # avoiding 0px images

    def fit_in_resize(self):
        source_width, source_height = self.engine.size

        # invert width and height if image orientation is not the
        # same as request orientation and need adaptive
        if self.context.request.adaptive and (
            (
                source_width < source_height
                and self.target_width > self.target_height
            )
            or (
                source_width > source_height
                and self.target_width < self.target_height
            )
        ):
            tmp = self.context.request.width
            self.context.request.width = self.context.request.height
            self.context.request.height = tmp
            tmp = self.target_width
            self.target_width = self.target_height
            self.target_height = tmp

        sign = 1
        if self.context.request.full:
            sign = -1

        if (
            sign == 1
            and self.target_width >= source_width
            and self.target_height >= source_height
        ):
            return

        # add 1e-5 to get .5 rounding up in Python3 as it was in Python2
        if (
            source_width / self.target_width * sign
            >= source_height / self.target_height * sign
        ):
            resize_height = round(
                source_height * self.target_width / source_width + 1e-5
            )
            resize_width = self.target_width
        else:
            resize_height = self.target_height
            resize_width = round(
                source_width * self.target_height / source_height + 1e-5
            )

        # ensure that filter should work on the real image size
        # and not on the request size which might be smaller than the
        # resized image in case `full-fit-in` is being used
        requested_width = (
            source_width
            if self.context.request.width == "orig"
            else self.context.request.width
        )
        requested_height = (
            source_height
            if self.context.request.height == "orig"
            else self.context.request.height
        )
        self.context.request.width = int(max(requested_width, resize_width))
        self.context.request.height = int(max(requested_height, resize_height))

        self.engine.resize(resize_width, resize_height)

    def debug(self):
        if not self.context.request.focal_points:
            return

        for point in self.context.request.focal_points:
            if point.width <= 1:
                point.width = 10
            if point.height <= 1:
                point.height = 10
            self.engine.draw_rectangle(
                point.x - (point.width // 2),
                point.y - (point.height // 2),
                point.width,
                point.height,
            )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from os.path import dirname
from shutil import which
from unittest.mock import Mock, patch

from preggy import expect
from tornado.testing import gen_test

from tests.fixtures.images import animated_image
from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer
from thumbor.result_storages.file_storage import Storage as FileResultStorage

JPEGTRAN_AVAILABLE = which("jpegtran") is not None
EXIFTOOL_AVAILABLE = which("exiftool") is not None


class ImageOperationsWithResultStorageTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path

        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path

        cfg.USE_GIFSICLE_ENGINE = True
        cfg.FFMPEG_PATH = which("ffmpeg")
        cfg.AUTO_WEBP = True
        cfg.OPTIMIZERS = [
            "thumbor.optimizers.gifv",
        ]

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")

        return ctx

    @property
    def result_storage(self):
        return self.context.modules.result_storage

    @patch("tornado.ioloop.IOLoop.instance")
    @gen_test
    async def test_saves_image_to_result_storage(self, instance_mock):
        instance_mock.return_value = self.io_loop
        response = await self.async_fetch(
            "/gTr2Xr9lbzIa2CT_dL_O0GByeR0=/animated.gif"
        )
        expect(response.code).to_equal(200)

        self.context.request = Mock(
            accepts_webp=False,
        )
        expected_path = self.result_storage.normalize_path(
            "/gTr2Xr9lbzIa2CT_dL_O0GByeR0=/animated.gif"
        )
        expect(expected_path).to_exist()
        expect(response.body).to_be_similar_to(animated_image())


class ImageOperationsResultStorageOnlyTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = "/path/that/does/not/exist"

        cfg.RESULT_STORAGE = "thumbor.result_storages.file_storage"
        cfg.RESULT_STORAGE_EXPIRATION_SECONDS = 60
        cfg.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path
        cfg.FFMPEG_PATH = which("ffmpeg")

        cfg.USE_GIFSICLE_ENGINE = True
        cfg.AUTO_WEBP = True
        cfg.OPTIMIZERS = [
            "thumbor.optimizers.gifv",
        ]

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")

        return ctx

    @property
    def result_storage(self):
        return self.context.modules.result_storage

    @gen_test
    async def test_loads_image_from_result_storage(self):
        self.context.request = Mock(
            accepts_webp=False,
        )
        expected_path = self.result_storage.normalize_path(
            "gTr2Xr9lbzIa2CT_dL_O0GByeR0=/animated.gif"
        )
        os.makedirs(dirname(expected_path))
        with open(expected_path, "wb") as img:
            img.write(animated_image())

        response = await self.async_fetch(
            "/gTr2Xr9lbzIa2CT_dL_O0GByeR0=/animated.gif"
        )
        expect(response.code).to_equal(200)
        expect(response.body).to_be_similar_to(animated_image())

    @patch.object(FileResultStorage, "get", side_effect=Exception)
    @gen_test
    async def test_loads_image_from_result_storage_fails_on_exception(self, _):
        response = await self.async_fetch(
            "/gTr2Xr9lbzIa2CT_dL_O0GByeR0=/animated.gif"
        )
        expect(response.code).to_equal(500)
        expect(response.body).to_be_empty()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.storages import BaseStorage


class Storage(BaseStorage):
    async def put(self, path, file_bytes):
        return path

    async def put_crypto(self, path):
        return path

    async def put_detector_data(self, path, data):
        return path

    async def get_crypto(self, path):
        return None

    async def get_detector_data(self, path):
        return None

    async def get(self, path):
        return None

    async def exists(self, path):
        return False

    async def remove(self, path):
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import unicodedata
from os.path import abspath, dirname, join


##
# Images used for tests :
#   - valid image      : JPEG 620x465, 69.88 KB
#   - too small image  : JPEG 20x20, 822 B
#   - too weight image : JPEG 300x400, 85.32 KB
##
def get_abs_path(img):
    return abspath(join(dirname(__file__), img))


VALID_IMAGE_PATH = get_abs_path("alabama1_ap620é.jpg")
SVG_IMAGE_PATH = get_abs_path("Commons-logo.svg")
TOO_SMALL_IMAGE_PATH = get_abs_path("20x20.jpg")
TOO_HEAVY_IMAGE_PATH = get_abs_path("Giunchedi%2C_Filippo_January_2015_01.jpg")
DEFAULT_IMAGE_PATH = get_abs_path("image.jpg")
ALABAMA1_IMAGE_PATH = get_abs_path("alabama1_ap620%C3%A9.jpg")
SPACE_IMAGE_PATH = get_abs_path("image%20space.jpg")
INVALID_QUANTIZATION_IMAGE_PATH = get_abs_path("invalid_quantization.jpg")
ANIMATED_IMAGE_PATH = get_abs_path("animated.gif")
NOT_SO_ANIMATED_IMAGE_PATH = get_abs_path("animated-one-frame.gif")


def get_image(img):
    encode_formats = ["NFD", "NFC", "NFKD", "NFKC"]
    for encode_format in encode_formats:
        try:
            with open(
                unicodedata.normalize(encode_format, img), "rb"
            ) as stream:
                body = stream.read()
                break
        except IOError:
            pass
    else:
        raise IOError(f"{img} not found")
    return body


def valid_image():
    return get_image(VALID_IMAGE_PATH)


def svg_image():
    return get_image(SVG_IMAGE_PATH)


def too_small_image():
    return get_image(TOO_SMALL_IMAGE_PATH)


def too_heavy_image():
    return get_image(TOO_HEAVY_IMAGE_PATH)


def default_image():
    return get_image(DEFAULT_IMAGE_PATH)


def alabama1():
    return get_image(ALABAMA1_IMAGE_PATH)


def space_image():
    return get_image(SPACE_IMAGE_PATH)


def invalid_quantization():
    return get_image(INVALID_QUANTIZATION_IMAGE_PATH)


def animated_image():
    return get_image(ANIMATED_IMAGE_PATH)


def not_so_animated_image():
    return get_image(NOT_SO_ANIMATED_IMAGE_PATH)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.compatibility import compatibility_get


def get_result(first, second, callback):
    callback(first, second, True)


class SimpleCompatibilityTestCase(TestCase):
    @gen_test
    async def test_can_get_result(self):
        first, second, response = await compatibility_get(
            1, second=2, func=get_result
        )
        expect(first).to_equal(1)
        expect(second).to_equal(2)
        expect(response).to_be_true()

    @gen_test
    async def test_fails_without_func(self):
        with expect.error_to_happen(
            RuntimeError,
            message="'func' argument can't be None when calling thumbor's compatibility layer.",
        ):
            await compatibility_get(1, second=2)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import logging
import logging.config
import os
import sys
import warnings
from os.path import dirname, expanduser
from shutil import which
from socket import socket

import tornado.ioloop
from PIL import Image
from tornado.httpserver import HTTPServer
from tornado.netutil import bind_unix_socket

from thumbor.config import Config
from thumbor.console import get_server_parameters
from thumbor.context import Context
from thumbor.importer import Importer
from thumbor.signal_handler import setup_signal_handler


def get_as_integer(value):
    try:
        return int(value)
    except (ValueError, TypeError):
        return None


def get_config(config_path, use_environment=False):
    if use_environment:
        Config.allow_environment_variables()

    lookup_paths = [os.curdir, expanduser("~"), "/etc/", dirname(__file__)]

    return Config.load(
        config_path, conf_name="thumbor.conf", lookup_paths=lookup_paths
    )


def configure_log(config, log_level):
    if config.THUMBOR_LOG_CONFIG and config.THUMBOR_LOG_CONFIG != "":
        logging.config.dictConfig(config.THUMBOR_LOG_CONFIG)
    else:
        logging.basicConfig(
            level=getattr(logging, log_level),
            format=config.THUMBOR_LOG_FORMAT,
            datefmt=config.THUMBOR_LOG_DATE_FORMAT,
        )


def get_importer(config):
    importer = Importer(config)
    importer.import_modules()

    if importer.error_handler_class is not None:
        importer.error_handler = (
            importer.error_handler_class(  # pylint: disable=not-callable
                config
            )
        )

    return importer


def validate_config(config, server_parameters):
    if server_parameters.security_key is None:
        server_parameters.security_key = config.SECURITY_KEY

    if not isinstance(server_parameters.security_key, (bytes, str)):
        raise RuntimeError(
            "No security key was found for this instance of thumbor. "
            + "Please provide one using the conf file or a security key file."
        )

    if config.ENGINE or config.USE_GIFSICLE_ENGINE:
        # Error on Image.open when image pixel count is above MAX_IMAGE_PIXELS
        warnings.simplefilter("error", Image.DecompressionBombWarning)

    if config.USE_GIFSICLE_ENGINE:
        server_parameters.gifsicle_path = which("gifsicle")

        if server_parameters.gifsicle_path is None:
            raise RuntimeError(
                "If using USE_GIFSICLE_ENGINE configuration to True,"
                " the `gifsicle` binary must be in the PATH "
                "and must be an executable."
            )


def get_context(server_parameters, config, importer):
    return Context(server=server_parameters, config=config, importer=importer)


def get_application(context):
    return context.modules.importer.import_class(context.app_class)(context)


def get_socket_from_fd(fname_or_fd, *, non_blocking=False):
    fd_number = get_as_integer(fname_or_fd)

    if fd_number is not None:
        sock = socket(fileno=fd_number)
        if non_blocking:
            sock.setblocking(False)
    else:
        sock = bind_unix_socket(fname_or_fd)

    return sock


def run_server(application, context):
    server = HTTPServer(application, xheaders=True)

    if context.server.fd is not None:
        sock = get_socket_from_fd(
            context.server.fd,
            non_blocking=context.config.NON_BLOCKING_SOCKETS,
        )
        server.add_socket(sock)

        logging.debug("thumbor starting at fd %s", context.server.fd)
    else:
        server.bind(context.server.port, context.server.ip)

        logging.debug(
            "thumbor starting at %s:%d", context.server.ip, context.server.port
        )

    server.start(context.server.processes)

    return server


def main(arguments=None):
    """Runs thumbor server with the specified arguments."""

    if arguments is None:
        arguments = sys.argv[1:]

    server_parameters = get_server_parameters(arguments)
    config = get_config(
        server_parameters.config_path, server_parameters.use_environment
    )
    configure_log(config, server_parameters.log_level.upper())

    validate_config(config, server_parameters)

    importer = get_importer(config)

    with get_context(server_parameters, config, importer) as context:
        application = get_application(context)
        server = run_server(application, context)
        setup_signal_handler(server, config)
        tornado.ioloop.IOLoop.instance().start()


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

# Test file
# pylint: disable=protected-access

import re
import time
from os.path import abspath, dirname, join
from urllib.parse import quote

import tornado.web
from preggy import expect
from tornado.httpclient import AsyncHTTPClient
from tornado.testing import gen_test

import thumbor.loaders.http_loader as loader
from tests.base import TestCase
from thumbor.config import Config
from thumbor.context import Context
from thumbor.loaders import LoaderResult


def fixture_for(filename):
    return abspath(join(dirname(__file__), "fixtures", filename))


class MainHandler(tornado.web.RequestHandler):
    async def get(self):
        self.write("Hello")


class TimeoutHandler(tornado.web.RequestHandler):
    async def get(self):
        time.sleep(1.2)
        self.write("Hello")


class EchoUserAgentHandler(tornado.web.RequestHandler):
    async def get(self):
        self.write(self.request.headers["User-Agent"])


class EchoAllHeadersHandler(tornado.web.RequestHandler):
    async def get(self):
        for header, value in sorted(self.request.headers.items()):
            self.write(f"{header}:{value}\n")


class HandlerMock:
    def __init__(self, headers):
        self.request = RequestMock(headers)


class RequestMock:
    def __init__(self, headers):
        self.headers = headers


class ResponseMock:
    def __init__(self, error=None, content_type=None, body=None, code=None):
        self.error = error
        self.code = code
        self.time_info = None

        self.headers = {"Content-Type": "image/jpeg"}

        if content_type:
            self.headers["Content-Type"] = content_type

        self.body = body


class ReturnContentTestCase(TestCase):
    def test_return_none_on_error(self):
        response_mock = ResponseMock(error="Error", code=599)
        ctx = Context(None, None, None)
        result = loader.return_contents(response_mock, "some-url", ctx)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_be_null()
        expect(result.successful).to_be_false()

    def test_return_body_if_valid(self):
        response_mock = ResponseMock(body="body", code=200)
        ctx = Context(None, None, None)
        result = loader.return_contents(response_mock, "some-url", ctx)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("body")

    def test_return_upstream_error_on_body_none(self):
        response_mock = ResponseMock(body=None, code=200)
        ctx = Context(None, None, None)
        result = loader.return_contents(response_mock, "some-url", ctx)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_be_null()
        expect(result.successful).to_be_false()
        expect(result.error).to_equal(LoaderResult.ERROR_UPSTREAM)

    def test_return_upstream_error_on_body_empty(self):
        response_mock = ResponseMock(body="", code=200)
        ctx = Context(None, None, None)
        result = loader.return_contents(response_mock, "some-url", ctx)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_be_null()
        expect(result.successful).to_be_false()
        expect(result.error).to_equal(LoaderResult.ERROR_UPSTREAM)


class ValidateUrlTestCase(TestCase):
    def test_with_allowed_sources(self):
        config = Config()
        config.ALLOWED_SOURCES = [
            "s.glbimg.com",
            re.compile(r"https://www\.google\.com/img/.*"),
        ]
        ctx = Context(None, config, None)
        expect(
            loader.validate(ctx, "http://www.google.com/logo.jpg")
        ).to_be_false()
        expect(
            loader.validate(ctx, "http://s2.glbimg.com/logo.jpg")
        ).to_be_false()
        expect(
            loader.validate(
                ctx,
                "/glob=:sfoir%20%20%3Co-pmb%20%20%20%20_%20%20%20%200%20%20g.-%3E%3Ca%20hplass=",  # NOQA, pylint: disable=line-too-long
            )
        ).to_be_false()
        expect(
            loader.validate(ctx, "https://www.google.com/img/logo.jpg")
        ).to_be_true()
        expect(
            loader.validate(ctx, "http://s.glbimg.com/logo.jpg")
        ).to_be_true()

    def test_without_allowed_sources(self):
        config = Config()
        config.ALLOWED_SOURCES = []
        ctx = Context(None, config, None)
        is_valid = loader.validate(ctx, "http://www.google.com/logo.jpg")
        expect(is_valid).to_be_true()


class NormalizeUrlTestCase(TestCase):
    def test_should_normalize_url(self):
        for url in ["http://some.url", "some.url"]:
            expect(loader._normalize_url(url)).to_equal("http://some.url")

    def test_should_normalize_quoted_url(self):
        url = (
            "https%3A//www.google.ca/images/branding/googlelogo/2x"
            "/googlelogo_color_272x92dp.png"
        )
        expected = (
            "https://www.google.ca/images/branding/googlelogo/2x"
            "/googlelogo_color_272x92dp.png"
        )
        result = loader._normalize_url(url)
        expect(result).to_equal(expected)


class DummyAsyncHttpClientTestCase(TestCase):
    # By default Tornado allows to create one (Async)HttpClient per
    # IOLoop instance
    # http://www.tornadoweb.org/en/stable/httpclient.html#tornado.httpclient.AsyncHTTPClient
    # AsyncHTTPTestCase provides a lot of useful code, which starts http
    # server listening with an app running.
    # But it also constructs AsyncHttpClient, which then by default is
    # getting reused upon next calls to get new AsyncHttpClient. Some
    # test cases are requiring curl client to be initialized
    # (see http_loader.py) but, by the time http_loader configures
    # Tornado's AsyncHTTPClient, it's already too late,
    # since Tornado has http client initialized for given IOLoop.
    # Forcing new instance here, on test start up time, is ensuring
    # that it won't be treated as singleton and rather be disposable instance.
    def get_http_client(self):
        return AsyncHTTPClient(force_instance=True)

    def tearDown(self):
        AsyncHTTPClient().close()  # clean up singleton instance
        super().tearDown()


class HttpLoaderTestCase(DummyAsyncHttpClientTestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", MainHandler)])

        return application

    @gen_test
    async def test_load_with_callback(self):
        url = self.get_url("/")
        config = Config()
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("Hello")
        expect(result.successful).to_be_true()

    @gen_test
    async def test_load_with_utf8_url(self):
        url = self.get_url(quote("/maracujá.jpg".encode("utf-8")))
        config = Config()
        ctx = Context(None, config, None)

        with expect.error_not_to_happen(UnicodeDecodeError):
            await loader.load(ctx, url)

    @gen_test
    async def test_load_with_curl(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT = True
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("Hello")
        expect(result.successful).to_be_true()


class HttpLoaderWithHeadersForwardingTestCase(DummyAsyncHttpClientTestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", EchoAllHeadersHandler)])

        return application

    @gen_test
    async def test_load_with_some_headers(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_HEADERS_WHITELIST = ["X-Server"]
        handler_mock_options = {
            "Accept-Encoding": "gzip",
            "User-Agent": "Thumbor",
            "Host": "localhost",
            "Accept": "*/*",
            "X-Server": "thumbor",
        }
        ctx = Context(None, config, None, HandlerMock(handler_mock_options))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer.decode()).to_include("X-Server:thumbor")

    @gen_test
    async def test_load_with_some_excluded_headers(self):
        url = self.get_url("/")
        config = Config()
        handler_mock_options = {
            "Accept-Encoding": "gzip",
            "User-Agent": "Thumbor",
            "Host": "localhost",
            "Accept": "*/*",
            "X-Server": "thumbor",
        }
        ctx = Context(None, config, None, HandlerMock(handler_mock_options))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer.decode()).Not.to_include("X-Server:thumbor")

    @gen_test
    async def test_load_with_all_headers(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_ALL_HEADERS = True
        handler_mock_options = {
            "X-Test": "123",
            "DNT": "1",
            "X-Server": "thumbor",
        }
        ctx = Context(None, config, None, HandlerMock(handler_mock_options))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer.decode()).to_include("Dnt:1\n")
        expect(result.buffer.decode()).to_include("X-Server:thumbor\n")
        expect(result.buffer.decode()).to_include("X-Test:123\n")

    @gen_test
    async def test_load_with_empty_accept(self):
        url = self.get_url("/")
        config = Config()
        handler_mock_options = {
            "Accept-Encoding": "gzip",
            "User-Agent": "Thumbor",
            "Host": "localhost",
            "Accept": "",
            "X-Server": "thumbor",
        }
        ctx = Context(None, config, None, HandlerMock(handler_mock_options))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer.decode()).to_include(
            "Accept:image/*;q=0.9,*/*;q=0.1\n"
        )


class HttpLoaderWithUserAgentForwardingTestCase(DummyAsyncHttpClientTestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", EchoUserAgentHandler)])

        return application

    @gen_test
    async def test_load_with_user_agent(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_USER_AGENT = True
        ctx = Context(
            None, config, None, HandlerMock({"User-Agent": "test-user-agent"})
        )

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("test-user-agent")

    @gen_test
    async def test_load_with_default_user_agent(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_FORWARD_USER_AGENT = True
        config.HTTP_LOADER_DEFAULT_USER_AGENT = "DEFAULT_USER_AGENT"
        ctx = Context(None, config, None, HandlerMock({}))

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("DEFAULT_USER_AGENT")


class HttpCurlTimeoutLoaderTestCase(DummyAsyncHttpClientTestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", TimeoutHandler)])

        return application

    @gen_test
    async def test_load_with_timeout(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT = True
        config.HTTP_LOADER_REQUEST_TIMEOUT = 1
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_be_null()
        expect(result.successful).to_be_false()

    @gen_test
    async def test_load_with_speed_timeout(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_CURL_ASYNC_HTTP_CLIENT = True
        config.HTTP_LOADER_CURL_LOW_SPEED_TIME = 1
        config.HTTP_LOADER_CURL_LOW_SPEED_LIMIT = 1000000000
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_be_null()
        expect(result.successful).to_be_false()


class HttpTimeoutLoaderTestCase(DummyAsyncHttpClientTestCase):
    def get_app(self):
        application = tornado.web.Application([(r"/", TimeoutHandler)])

        return application

    @gen_test
    async def test_load_without_curl_but_speed_timeout(self):
        url = self.get_url("/")
        config = Config()
        config.HTTP_LOADER_CURL_LOW_SPEED_TIME = 1
        config.HTTP_LOADER_CURL_LOW_SPEED_LIMIT = 1000000000
        ctx = Context(None, config, None)

        result = await loader.load(ctx, url)
        expect(result).to_be_instance_of(LoaderResult)
        expect(result.buffer).to_equal("Hello")
        expect(result.successful).to_be_true()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath, dirname, isabs, join
from typing import Dict

import numpy as np

from thumbor.detectors import BaseDetector
from thumbor.point import FocalPoint


class CascadeLoaderDetector(BaseDetector):
    def load_cascade_file(self, module_path, cascade_file_path):
        import cv2  # noqa pylint: disable=import-error,import-outside-toplevel

        if not hasattr(self.__class__, "cascade"):
            if isabs(cascade_file_path):
                cascade_file = cascade_file_path
            else:
                cascade_file = join(
                    abspath(dirname(module_path)), cascade_file_path
                )
            self.__class__.cascade = (
                cv2.CascadeClassifier(  # pylint: disable=no-member
                    cascade_file
                )
            )

    def get_min_size_for(self, size):
        ratio = int(min(size) / 15)
        ratio = max(20, ratio)

        return (ratio, ratio)

    def get_features(self):
        engine = self.context.modules.engine
        img = np.array(
            engine.convert_to_grayscale(update_image=False, alpha=False)
        )

        faces = self.__class__.cascade.detectMultiScale(
            img, 1.2, 4, minSize=self.get_min_size_for(engine.size)
        )
        faces_scaled = []

        for (posx, posy, width, height) in faces:
            faces_scaled.append(
                ((posx.item(), posy.item(), width.item(), height.item()), 0)
            )

        return faces_scaled

    def get_origin(self) -> str:
        return "detection"

    def get_detection_offset(
        self,
        left: int,  # pylint: disable=unused-argument
        top: int,  # pylint: disable=unused-argument
        width: int,  # pylint: disable=unused-argument
        height: int,  # pylint: disable=unused-argument
    ) -> Dict[str, int]:
        return {}

    async def detect(self):
        if not self.verify_cv():
            await self.next()

            return

        features = self.get_features()

        if not features:
            await self.next()  # pylint: disable=not-callable

        for (left, top, width, height), _ in features:
            offset = self.get_detection_offset(left, top, width, height)
            self.context.request.focal_points.append(
                FocalPoint.from_square(
                    left + offset.get("left", 0.0),
                    top + offset.get("top", 0.0),
                    width + offset.get("right", 0.0),
                    height + offset.get("bottom", 0.0),
                    origin=self.get_origin(),
                )
            )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.handlers import BaseHandler


class HealthcheckHandler(BaseHandler):
    async def get(self):
        self.set_header("Cache-Control", "no-cache")
        self.write("WORKING")

    async def head(self):
        self.set_header("Cache-Control", "no-cache")
        self.set_status(200)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

# pylint: disable=protected-access

from time import sleep
from unittest import mock

import pytest
from preggy import expect

from thumbor.threadpool import ThreadPool


@pytest.mark.asyncio
async def test_can_get_threadpool_instance():
    instance = ThreadPool.instance(0)
    expect(instance.pool).to_be_null()

    instance = ThreadPool.instance(10)
    expect(instance).not_to_be_null()

    instance2 = ThreadPool.instance(10)
    expect(instance2).to_equal(instance)

    instance3 = ThreadPool.instance(11)
    expect(instance3).not_to_equal(instance)


@pytest.mark.asyncio
async def test_can_run_task_in_foreground():
    instance = ThreadPool.instance(0)
    expect(instance).not_to_be_null()

    def add():
        return 10

    result = await instance._execute_in_foreground(add)
    expect(result).to_equal(10)


@pytest.mark.asyncio
async def test_can_run_task_in_foreground_and_exception_happens():
    instance = ThreadPool.instance(0)
    expect(instance).not_to_be_null()
    exception = Exception("Boom")

    def add():
        raise exception

    with expect.error_to_happen(Exception, message="Boom"):
        await instance._execute_in_foreground(add)


@pytest.mark.asyncio
async def test_queueing_task_when_no_pool_runs_sync():
    instance = ThreadPool.instance(0)
    expect(instance).not_to_be_null()

    def add():
        return 10

    result = await instance.queue(add)
    expect(result).to_equal(10)


@pytest.mark.asyncio
async def test_queueing_task_when_no_pool_runs_sync_and_exception_happens():
    instance = ThreadPool.instance(0)
    expect(instance).not_to_be_null()
    exception = Exception("Boom")

    def add():
        raise exception

    with expect.error_to_happen(Exception, message="Boom"):
        await instance.queue(add)


@pytest.mark.asyncio
async def test_can_run_async():
    instance = ThreadPool.instance(10)
    expect(instance).not_to_be_null()

    def add():
        sleep(1)
        return 10

    result = await instance._execute_in_pool(add)
    expect(result).to_equal(10)


@pytest.mark.asyncio
async def test_can_run_async_with_queue():
    instance = ThreadPool.instance(10)
    expect(instance).not_to_be_null()

    def add(value):
        sleep(1)
        return value

    result = await instance.queue(add, 10)
    expect(result).to_equal(10)


@pytest.mark.asyncio
async def test_can_cleanup_pool():
    instance = ThreadPool.instance(0)
    instance.pool = mock.Mock()
    instance.cleanup()

    expect(instance.pool.shutdown.called).to_be_true()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from libthumbor.url import Url
from preggy import expect

from thumbor.app import ThumborServiceApp
from thumbor.testing import TestCase


class AppTestCase(TestCase):
    def test_can_create_app(self):
        app = ThumborServiceApp(self.context)
        expect(app).not_to_be_null()
        expect(app.context).to_equal(self.context)

    def test_can_get_handlers(self):
        ctx = self.get_context()
        ctx.config.UPLOAD_ENABLED = False
        ctx.config.USE_BLACKLIST = False
        ctx.config.HEALTHCHECK_ROUTE = "/health"
        app = ThumborServiceApp(ctx)

        handlers = app.get_handlers()
        expect(handlers).to_length(2)
        expect(handlers[0][0]).to_equal(r"/health")
        expect(handlers[1][0]).to_equal(Url.regex())

    def test_can_get_handlers_with_upload(self):
        ctx = self.get_context()
        ctx.config.UPLOAD_ENABLED = True
        ctx.config.USE_BLACKLIST = False
        app = ThumborServiceApp(ctx)

        handlers = app.get_handlers()
        expect(handlers).to_length(4)

    def test_can_get_handlers_with_blacklist(self):
        ctx = self.get_context()
        ctx.config.UPLOAD_ENABLED = False
        ctx.config.USE_BLACKLIST = True
        app = ThumborServiceApp(ctx)

        handlers = app.get_handlers()
        expect(handlers).to_length(3)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.compatibility import compatibility_get


async def load(context, path):
    loader = context.modules.compatibility_legacy_loader
    if loader is None:
        raise RuntimeError(
            "The 'COMPATIBILITY_LEGACY_LOADER' configuration should point "
            "to a valid loader when using compatibility loader."
        )

    return await compatibility_get(context, path, func=loader.load)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class FocalFilterTestCase(FilterTestCase):
    @gen_test
    async def test_focal_filter_valid(self):
        await self.get_filtered(
            "source.jpg", "thumbor.filters.focal", "focal(146x156:279x208)"
        )

        expect(self.context.request.focal_points[0].origin).to_equal(
            "Explicit"
        )
        expect(self.context.request.focal_points[0].height).to_equal(52)
        expect(self.context.request.focal_points[0].width).to_equal(133)
        expect(self.context.request.focal_points[0].y).to_equal(182)
        expect(self.context.request.focal_points[0].x).to_equal(212)

    @gen_test
    async def test_focal_filter_no_change(self):
        image = await self.get_filtered(
            "source.jpg",
            "thumbor.filters.focal",
            "focal(0x0:400x600)",
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_focal_filter_pass_through_empty(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.focal", "focal()"
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_focal_filter_pass_through_invalid(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.focal", "focal(0x0:0x0)"
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_focal_filter_pass_through_dumb(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.focal", "focal(lkasjdklfjlkajsfd)"
        )
        expected = self.get_fixture("source.jpg")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2015 popsugar.com vtodorov@popsugar.com

from thumbor.result_storages import BaseStorage


class Storage(BaseStorage):
    async def put(self, image_bytes):
        return ""

    async def get(self):
        return None
